from typing import Any, Optional
import weakref

from neuralogic.core.settings.settings_proxy import SettingsProxy
from neuralogic.core.enums import Optimizer, Initializer, ErrorFunction, Activation


class Settings:
    def __init__(
        self,
        *,
        optimizer: Optimizer = Optimizer.ADAM,
        learning_rate: Optional[float] = None,
        epochs: int = 3000,
        error_function: ErrorFunction = ErrorFunction.SQUARED_DIFF,
        initializer: Initializer = Initializer.UNIFORM,
        initializer_const: float = 0.1,
        initializer_uniform_scale: float = 2.0,
        rule_neuron_activation: Activation = Activation.TANH,
        relation_neuron_activation: Activation = Activation.TANH,
        iso_value_compression: bool = True,
        chain_pruning: bool = True,
    ):
        self.params = locals().copy()
        self.params.pop("self")
        self._proxies = weakref.WeakSet()

        if learning_rate is None:
            self.params["learning_rate"] = 0.1 if optimizer == Optimizer.SGD else 0.001

    @property
    def iso_value_compression(self) -> bool:
        return self.params["iso_value_compression"]

    @iso_value_compression.setter
    def iso_value_compression(self, iso_value_compression: bool):
        self._update("iso_value_compression", iso_value_compression)

    @property
    def chain_pruning(self) -> bool:
        return self.params["chain_pruning"]

    @chain_pruning.setter
    def chain_pruning(self, chain_pruning: bool):
        self._update("chain_pruning", chain_pruning)

    @property
    def seed(self) -> int:
        return self.params["seed"]

    @seed.setter
    def seed(self, seed: int):
        self._update("seed", seed)

    @property
    def learning_rate(self) -> float:
        return self.params["learning_rate"]

    @learning_rate.setter
    def learning_rate(self, learning_rate: float):
        self._update("learning_rate", learning_rate)

    @property
    def optimizer(self) -> Optimizer:
        return self.params["optimizer"]

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer):
        self._update("optimizer", optimizer)

    @property
    def initializer_const(self) -> float:
        return self.params["initializer_const"]

    @initializer_const.setter
    def initializer_const(self, value: float):
        self._update("initializer_const", value)

    @property
    def initializer_uniform_scale(self) -> float:
        return self.params["initializer_uniform_scale"]

    @initializer_uniform_scale.setter
    def initializer_uniform_scale(self, value: float):
        self._update("initializer_uniform_scale", value)

    @property
    def error_function(self) -> ErrorFunction:
        return self.params["error_function"]

    @error_function.setter
    def error_function(self, error_function: ErrorFunction):
        self._update("error_function", error_function)

    @property
    def epochs(self) -> int:
        return self.params["epochs"]

    @epochs.setter
    def epochs(self, epochs: int):
        self._update("epochs", epochs)

    @property
    def initializer(self) -> Initializer:
        return self.params["initializer"]

    @initializer.setter
    def initializer(self, initializer: Initializer):
        self._update("initializer", initializer)

    @property
    def relation_neuron_activation(self) -> Activation:
        return self.params["relation_neuron_activation"]

    @relation_neuron_activation.setter
    def relation_neuron_activation(self, value: Activation):
        self._update("relation_neuron_activation", value)

    @property
    def rule_neuron_activation(self) -> Activation:
        return self.params["rule_neuron_activation"]

    @rule_neuron_activation.setter
    def rule_neuron_activation(self, value: Activation):
        self._update("rule_neuron_activation", value)

    def create_proxy(self) -> SettingsProxy:
        proxy = SettingsProxy(**self.params)
        self._proxies.add(proxy)

        return proxy

    def create_disconnected_proxy(self) -> SettingsProxy:
        return SettingsProxy(**self.params)

    def _update(self, parameter: str, value: Any) -> None:
        if parameter not in self.params:
            raise NotImplementedError
        self.params[parameter] = value

        for proxy in self._proxies.copy():
            proxy.__setattr__(parameter, value)



import os
from typing import Optional

import jpype


_is_initialized = False
_std_out = None
_std_err = None

jvm_params = {
    "classpath": os.path.join(os.path.abspath(os.path.dirname(__file__)), "jar", "NeuraLogic.jar"),
}


class TextIOWrapper:
    def __init__(self, wrapped_text_io):
        self.wrapped_text_io = wrapped_text_io

    def write(self, string):
        self.wrapped_text_io.write(str(string))


def set_system_output(output, system_output_setter) -> None:
    java_output_stream = os.devnull

    if output is not None:
        wrapped = TextIOWrapper(output)

        java_io_wrapper = jpype.JProxy("cz.cvut.fel.ida.utils.python.PythonOutputStream.TextIOWrapper", inst=wrapped)
        java_output_stream = jpype.JClass("cz.cvut.fel.ida.utils.python.PythonOutputStream")(java_io_wrapper)
    system_output_setter(jpype.java.io.PrintStream(java_output_stream))


def set_stdout(out_io=None) -> None:
    global _std_out
    _std_out = out_io

    if not _is_initialized:
        return

    set_system_output(_std_out, jpype.java.lang.System.setOut)


def set_stderr(err_io=None) -> None:
    global _std_err
    _std_err = err_io

    if not _is_initialized:
        return

    set_system_output(_std_err, jpype.java.lang.System.setErr)


def set_jvm_path(path: Optional[str]) -> None:
    global jvm_params

    if path is None:
        jvm_params.pop("jvmpath", None)
    else:
        jvm_params["jvmpath"] = path


def is_initialized() -> bool:
    return _is_initialized


def initialize(
    debug_mode: bool = False, debug_port: int = 12999, is_debug_server: bool = True, debug_suspend: bool = False
):
    global _is_initialized

    if _is_initialized:
        raise Exception("NeuraLogic already initialized")

    _is_initialized = True
    if debug_mode:
        port = int(debug_port)
        server = "y" if is_debug_server else "n"
        suspend = "y" if debug_suspend else "n"

        debug_params = [
            "-Xint",
            "-Xdebug",
            "-Xnoagent",
            f"-Xrunjdwp:transport=dt_socket,server={server},address={port},suspend={suspend}",
        ]

        jpype.startJVM(*debug_params, **jvm_params)
    else:
        jpype.startJVM(**jvm_params)

    set_stderr(_std_err)
    set_stdout(_std_out)



from typing import List, Optional, Dict
import torch
import numpy as np

from neuralogic.core.builder import Sample, Weight, Neuron
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.enums import Initializer, Backend
from neuralogic.nn.base import AbstractNeuraLogic


def longtail(tensor: torch.Tensor, _: SettingsProxy):
    init = np.random.rand(*tensor.size())

    x0, x1, power = 0.0, 10.0, 50.0
    x0_power = np.power(x0, power + 1)

    init = x1 - np.power((np.power(x1, power + 1) - x0_power) * init + x0_power, 1 / (power + 1))
    tensor.fill_(init)


class NeuraLogic(AbstractNeuraLogic):
    activations = {
        "Sigmoid": torch.sigmoid,
        "Average": torch.mean,
        "Maximum": torch.max,
        "ReLu": torch.relu,
        "Tanh": torch.tanh,
    }

    initializers = {
        Initializer.NORMAL: lambda tensor, _: torch.nn.init.normal_(tensor),
        Initializer.UNIFORM: lambda tensor, settings: torch.nn.init.uniform_(
            tensor, -settings.initializer_uniform_scale / 2, settings.initializer_uniform_scale / 2
        ),
        Initializer.CONSTANT: lambda tensor, settings: torch.nn.init.constant_(tensor, settings.initializer_const),
        Initializer.LONGTAIL: longtail,
        # TODO: GLOROT and HE
    }

    def __init__(self, model: List[Weight], template, settings: Optional[SettingsProxy] = None):
        super().__init__(Backend.TORCH, template, settings)

        self.model: Optional[torch.nn.ParameterList] = None
        self.weights_meta = model
        self.weights: List[torch.nn.Parameter] = []

        self.reset_parameters()

    def reset_parameters(self):
        initializer = Initializer[str(self.settings.initializer)]

        if initializer not in NeuraLogic.initializers:
            raise NotImplementedError

        weight_initializer = NeuraLogic.initializers[initializer]

        for weight in self.weights_meta:
            if weight.fixed:
                parameter = torch.nn.Parameter(torch.tensor(weight.value, dtype=torch.float64), requires_grad=False)
            else:
                parameter = torch.nn.Parameter(
                    torch.empty(size=weight.dimensions, requires_grad=True, dtype=torch.float64)
                )
                weight_initializer(parameter.data, self.settings)
            self.weights.append(parameter)
        self.model = torch.nn.ParameterList(self.weights)

    def build_sample(self, sample: Sample):
        self.hooks_set = len(self.hooks) != 0

        torch_neurons: List[Optional[torch.Tensor]] = [None] * len(sample.neurons)

        for neuron in sample.neurons:
            torch_neurons[neuron.index] = self.to_torch_expression(neuron, torch_neurons, self.model)
        return torch_neurons[sample.neurons[-1].index]

    def __call__(self, sample: Sample):
        return self.build_sample(sample)

    def state_dict(self) -> Dict:
        weights = {}

        for meta, weight in zip(self.weights_meta, self.weights):
            if not meta.fixed and meta.index >= 0:
                weights[meta.index] = weight if isinstance(weight, (int, float)) else weight.data
        return {"weights": weights}

    def load_state_dict(self, state_dict: Dict):
        weight_dict = state_dict["weights"]

        for i, (meta, weight) in enumerate(zip(self.weights_meta, self.weights)):
            if meta.index < 0 or meta.fixed:
                continue

            if isinstance(weight, (int, float)):
                weight.data = torch.tensor([weight_dict[meta.index]], dtype=torch.float64)
            else:
                weight.data = torch.tensor(weight_dict[meta.index], dtype=torch.float64)

    @staticmethod
    def to_tensor_value(value) -> torch.Tensor:
        if isinstance(value, (float, int)):
            return torch.tensor([value], requires_grad=False, dtype=torch.float64)
        return torch.tensor(value, requires_grad=False, dtype=torch.float64)

    def to_torch_expression(
        self, neuron: Neuron, neurons: List[torch.Tensor], weights: torch.nn.ParameterList
    ) -> torch.Tensor:
        if neuron.inputs:
            out = self.process_neuron_inputs(neuron, neurons, weights)

            if neuron.activation:
                if not neuron.pooling:
                    out = sum(out)
                else:
                    out = torch.stack(out)
                if neuron.activation != "Identity":
                    out = NeuraLogic.activations[neuron.activation](out)
            else:
                out = sum(out)
        else:
            out = NeuraLogic.to_tensor_value(neuron.value)

            if neuron.activation and neuron.activation != "Identity":
                out = NeuraLogic.activations[neuron.activation](out)

        if self.hooks_set and neuron.hook_name is not None and neuron.hook_name in self.hooks:
            self.run_hook(neuron.hook_name, out.value())
        return out

    def process_neuron_inputs(
        self, neuron: Neuron, neurons: List[torch.Tensor], weights: torch.nn.ParameterList
    ) -> List[torch.Tensor]:
        out = []

        if neuron.weights:
            for w, i in zip(neuron.weights, neuron.inputs):
                neuron = neurons[i]
                weight = weights[w]

                neuron_size = neuron.size()
                weight_size = weight.size()

                if not neuron_size or not weight_size or neuron_size == (1,) or weight_size == (1,):
                    out.append(torch.multiply(weight, neuron))
                else:
                    out.append(torch.matmul(weight, neuron))
        else:
            for i in neuron.inputs:
                out.append(neurons[i])
        return out



from typing import List, Union, Optional

import jpype

from neuralogic import is_initialized, initialize
from neuralogic.core import Template, JavaFactory, Settings
from neuralogic.core.builder import DatasetBuilder
from neuralogic.core.constructs.atom import AtomType
from neuralogic.core.constructs.rule import Rule


class InferenceEngine:
    def __init__(self, template: Template):
        if not is_initialized():
            initialize()

        self.settings = Settings().create_disconnected_proxy()
        self.java_factory = JavaFactory()

        self.settings.settings.inferTemplateFacts = False

        self.parsed_template = template.get_parsed_template(self.settings, self.java_factory)
        self.dataset_builder = DatasetBuilder(self.parsed_template, self.java_factory)

        self.examples: List[Union[AtomType, Rule]] = []

        self.grounder = jpype.JClass("cz.cvut.fel.ida.logic.grounding.Grounder").getGrounder(self.settings.settings)
        field = self.grounder.getClass().getDeclaredField("herbrandModel")
        field.setAccessible(True)

        self.herbrand_model = field.get(self.grounder)

        self.examples_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.ExamplesBuilder")
        self.queries_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.QueriesBuilder")
        self.grounding_sample = jpype.JClass("cz.cvut.fel.ida.logic.grounding.GroundingSample")
        self.horn_clause = jpype.JClass("cz.cvut.fel.ida.logic.HornClause")

        self.empty_example = jpype.JClass("cz.cvut.fel.ida.logic.constructs.example.LiftedExample")()

    def set_knowledge(self, examples: List[Union[AtomType, Rule]]) -> None:
        self.examples = examples

    def q(self, query: AtomType, examples: Optional[List[Union[AtomType, Rule]]] = None):
        return self.query(query, examples)

    def query(self, query: AtomType, examples: Optional[List[Union[AtomType, Rule]]] = None):
        if examples is None:
            examples = self.examples

        examples_builder = self.examples_builder(self.settings.settings)
        query_builder = self.queries_builder(self.settings.settings)
        query_builder.setFactoriesFrom(examples_builder)

        self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
        queries = self.dataset_builder.build_queries([query], query_builder)

        self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
        examples = jpype.java.util.ArrayList(self.dataset_builder.build_examples([examples], examples_builder)[0])

        logic_samples = DatasetBuilder.merge_queries_with_examples(queries, examples)
        logic_samples = jpype.java.util.ArrayList(logic_samples)

        sample = logic_samples[0]

        gs = self.grounding_sample(sample, self.parsed_template)

        lifted_example = gs.query.evidence
        template = gs.template

        ground_template = self.grounder.groundRulesAndFacts(lifted_example, template)

        clause = self.java_factory.atom_to_clause(query)
        horn_clause = self.horn_clause(clause)
        substitutions = self.herbrand_model.groundingSubstitutions(horn_clause)

        labels = list(substitutions.r)
        substitutions_sets = list(substitutions.s)

        def generator():
            for substitution_set in substitutions_sets:
                yield {str(label): str(substitution) for label, substitution in zip(labels, substitution_set)}

        if len(substitutions_sets) == 0:
            return {}
        return generator()
